@@ -3,15 +3,18 @@
 import struct
 import decimal
 import datetime
+import json
 
 from pymysql.util import byte2int
+from pymysql.charset import charset_by_name
 
 from .event import BinLogEvent
+from .exceptions import TableMetadataUnavailableError
 from .constants import FIELD_TYPE
 from .constants import BINLOG
 from .column import Column
 from .table import Table
-
+from .bitmap import BitCount, BitGet
 
 class RowsEvent(BinLogEvent):
     def __init__(self, from_packet, event_size, table_map, ctl_connection, **kwargs):
@@ -19,22 +22,36 @@ class RowsEvent(BinLogEvent):
                                         ctl_connection, **kwargs)
         self.__rows = None
         self.__only_tables = kwargs["only_tables"]
+        self.__ignored_tables = kwargs["ignored_tables"]
         self.__only_schemas = kwargs["only_schemas"]
+        self.__ignored_schemas = kwargs["ignored_schemas"]
 
         #Header
         self.table_id = self._read_table_id()
-        self.primary_key = table_map[self.table_id].data["primary_key"]
 
         # Additional information
-        self.schema = self.table_map[self.table_id].schema
-        self.table = self.table_map[self.table_id].table
+        try:
+            self.primary_key = table_map[self.table_id].data["primary_key"]
+            self.schema = self.table_map[self.table_id].schema
+            self.table = self.table_map[self.table_id].table
+        except KeyError: #If we have filter the corresponding TableMap Event
+            self._processed = False
+            return
 
         if self.__only_tables is not None and self.table not in self.__only_tables:
             self._processed = False
             return
+        elif self.__ignored_tables is not None and self.table in self.__ignored_tables:
+            self._processed = False
+            return
+
         if self.__only_schemas is not None and self.schema not in self.__only_schemas:
             self._processed = False
             return
+        elif self.__ignored_schemas is not None and self.schema in self.__ignored_schemas:
+            self._processed = False
+            return
+
 
         #Event V2
         if self.event_type == BINLOG.WRITE_ROWS_EVENT_V2 or \
@@ -49,24 +66,39 @@ class RowsEvent(BinLogEvent):
         self.number_of_columns = self.packet.read_length_coded_binary()
         self.columns = self.table_map[self.table_id].columns
 
+        if len(self.columns) == 0:  # could not read the table metadata, probably already dropped
+            self.complete = False
+            if self._fail_on_table_metadata_unavailable:
+                raise TableMetadataUnavailableError(self.table)
+
     def __is_null(self, null_bitmap, position):
         bit = null_bitmap[int(position / 8)]
         if type(bit) is str:
             bit = ord(bit)
         return bit & (1 << (position % 8))
 
-    def _read_column_data(self, null_bitmap):
+    def _read_column_data(self,  cols_bitmap):
         """Use for WRITE, UPDATE and DELETE events.
         Return an array of column data
         """
         values = {}
 
+        # null bitmap length = (bits set in 'columns-present-bitmap'+7)/8
+        # See http://dev.mysql.com/doc/internals/en/rows-event.html
+        null_bitmap = self.packet.read((BitCount(cols_bitmap) + 7) / 8)
+
+        nullBitmapIndex = 0
         nb_columns = len(self.columns)
         for i in range(0, nb_columns):
             column = self.columns[i]
             name = self.table_map[self.table_id].columns[i].name
             unsigned = self.table_map[self.table_id].columns[i].unsigned
-            if self.__is_null(null_bitmap, i):
+
+            if BitGet(cols_bitmap, i) == 0:
+                values[name] = None
+                continue
+
+            if self.__is_null(null_bitmap, nullBitmapIndex):
                 values[name] = None
             elif column.type == FIELD_TYPE.TINY:
                 if unsigned:
@@ -130,7 +162,7 @@ class RowsEvent(BinLogEvent):
                 values[name] = self.packet.read_uint8() + 1900
             elif column.type == FIELD_TYPE.ENUM:
                 values[name] = column.enum_values[
-                    self.packet.read_uint_by_size(column.size) - 1]
+                    self.packet.read_uint_by_size(column.size)]
             elif column.type == FIELD_TYPE.SET:
                 # We read set columns as a bitmap telling us which options
                 # are enabled
@@ -145,9 +177,14 @@ class RowsEvent(BinLogEvent):
             elif column.type == FIELD_TYPE.GEOMETRY:
                 values[name] = self.packet.read_length_coded_pascal_string(
                     column.length_size)
+            elif column.type == FIELD_TYPE.JSON:
+                values[name] = self.packet.read_binary_json(column.length_size)
             else:
                 raise NotImplementedError("Unknown MySQL column type: %d" %
                                           (column.type))
+
+            nullBitmapIndex += 1
+
         return values
 
     def __add_fsp_to_time(self, time, column):
@@ -155,6 +192,12 @@ class RowsEvent(BinLogEvent):
         For more details about new date format:
         http://dev.mysql.com/doc/internals/en/date-and-time-data-type-representation.html
         """
+        microsecond = self.__read_fsp(column)
+        if microsecond > 0:
+            time = time.replace(microsecond=microsecond)
+        return time
+
+    def __read_fsp(self, column):
         read = 0
         if column.fsp == 1 or column.fsp == 2:
             read = 1
@@ -165,15 +208,20 @@ class RowsEvent(BinLogEvent):
         if read > 0:
             microsecond = self.packet.read_int_be_by_size(read)
             if column.fsp % 2:
-                time = time.replace(microsecond=int(microsecond / 10))
-            else:
-                time = time.replace(microsecond=microsecond)
-        return time
+                microsecond = int(microsecond / 10)
+            return microsecond * (10 ** (6-column.fsp))
+        return 0
+
+    @staticmethod
+    def charset_to_encoding(name):
+        charset = charset_by_name(name)
+        return charset.encoding if charset else name
 
     def __read_string(self, size, column):
         string = self.packet.read_length_coded_pascal_string(size)
         if column.character_set_name is not None:
-            string = string.decode(column.character_set_name)
+            encoding = self.charset_to_encoding(column.character_set_name)
+            string = string.decode(encoding)
         return string
 
     def __read_bit(self, column):
@@ -201,10 +249,10 @@ class RowsEvent(BinLogEvent):
 
     def __read_time(self):
         time = self.packet.read_uint24()
-        date = datetime.time(
-            hour=int(time / 10000),
-            minute=int((time % 10000) / 100),
-            second=int(time % 100))
+        date = datetime.timedelta(
+            hours=int(time / 10000),
+            minutes=int((time % 10000) / 100),
+            seconds=int(time % 100))
         return date
 
     def __read_time2(self, column):
@@ -219,11 +267,20 @@ class RowsEvent(BinLogEvent):
         24 bits = 3 bytes
         """
         data = self.packet.read_int_be_by_size(3)
-        t = datetime.time(
-            hour=self.__read_binary_slice(data, 2, 10, 24),
-            minute=self.__read_binary_slice(data, 12, 6, 24),
-            second=self.__read_binary_slice(data, 18, 6, 24))
-        return self.__add_fsp_to_time(t, column)
+
+        sign = 1 if self.__read_binary_slice(data, 0, 1, 24) else -1
+        if sign == -1:
+            # negative integers are stored as 2's compliment
+            # hence take 2's compliment again to get the right value.
+            data = ~data + 1
+
+        t = datetime.timedelta(
+            hours=sign*self.__read_binary_slice(data, 2, 10, 24),
+            minutes=self.__read_binary_slice(data, 12, 6, 24),
+            seconds=self.__read_binary_slice(data, 18, 6, 24),
+            microseconds=self.__read_fsp(column)
+        )
+        return t
 
     def __read_date(self):
         time = self.packet.read_uint24()
@@ -231,11 +288,10 @@ class RowsEvent(BinLogEvent):
             return None
 
         year = (time & ((1 << 15) - 1) << 9) >> 9
-        if year == 0:
-            return None
-
         month = (time & ((1 << 4) - 1) << 5) >> 5
         day = (time & ((1 << 5) - 1))
+        if year == 0 or month == 0 or day == 0:
+            return None
 
         date = datetime.date(
             year=year,
@@ -290,6 +346,7 @@ class RowsEvent(BinLogEvent):
                 minute=self.__read_binary_slice(data, 28, 6, 40),
                 second=self.__read_binary_slice(data, 34, 6, 40))
         except ValueError:
+            self.__read_fsp(column)
             return None
         return self.__add_fsp_to_time(t, column)
 
@@ -363,14 +420,18 @@ class RowsEvent(BinLogEvent):
 
     def _fetch_rows(self):
         self.__rows = []
-        while self.packet.read_bytes + 1 < self.event_size:
+
+        if not self.complete:
+            return
+
+        while self.packet.read_bytes < self.event_size:
             self.__rows.append(self._fetch_one_row())
 
-    def __getattr__(self, name):
-        if name == "rows":
-            if self.__rows is None:
-                self._fetch_rows()
-            return self.__rows
+    @property
+    def rows(self):
+        if self.__rows is None:
+            self._fetch_rows()
+        return self.__rows
 
 
 class DeleteRowsEvent(RowsEvent):
@@ -389,8 +450,7 @@ class DeleteRowsEvent(RowsEvent):
     def _fetch_one_row(self):
         row = {}
 
-        null_bitmap = self.packet.read((self.number_of_columns + 7) / 8)
-        row["values"] = self._read_column_data(null_bitmap)
+        row["values"] = self._read_column_data(self.columns_present_bitmap)
         return row
 
     def _dump(self):
@@ -418,8 +478,7 @@ class WriteRowsEvent(RowsEvent):
     def _fetch_one_row(self):
         row = {}
 
-        null_bitmap = self.packet.read((self.number_of_columns + 7) / 8)
-        row["values"] = self._read_column_data(null_bitmap)
+        row["values"] = self._read_column_data(self.columns_present_bitmap)
         return row
 
     def _dump(self):
@@ -454,12 +513,10 @@ class UpdateRowsEvent(RowsEvent):
 
     def _fetch_one_row(self):
         row = {}
-        null_bitmap = self.packet.read((self.number_of_columns + 7) / 8)
 
-        row["before_values"] = self._read_column_data(null_bitmap)
+        row["before_values"] = self._read_column_data(self.columns_present_bitmap)
 
-        null_bitmap = self.packet.read((self.number_of_columns + 7) / 8)
-        row["after_values"] = self._read_column_data(null_bitmap)
+        row["after_values"] = self._read_column_data(self.columns_present_bitmap2)
         return row
 
     def _dump(self):
@@ -476,16 +533,26 @@ class UpdateRowsEvent(RowsEvent):
 
 class TableMapEvent(BinLogEvent):
     """This evenement describe the structure of a table.
-    It's send before a change append on a table.
-    A end user of the lib should have no usage of this
+    It's sent before a change happens on a table.
+    An end user of the lib should have no usage of this
     """
 
     def __init__(self, from_packet, event_size, table_map, ctl_connection, **kwargs):
         super(TableMapEvent, self).__init__(from_packet, event_size,
                                             table_map, ctl_connection, **kwargs)
+        self.__only_tables = kwargs["only_tables"]
+        self.__ignored_tables = kwargs["ignored_tables"]
+        self.__only_schemas = kwargs["only_schemas"]
+        self.__ignored_schemas = kwargs["ignored_schemas"]
+        self.__freeze_schema = kwargs["freeze_schema"]
 
         # Post-Header
         self.table_id = self._read_table_id()
+
+        if self.table_id in table_map and self.__freeze_schema:
+            self._processed = False
+            return
+
         self.flags = struct.unpack('<H', self.packet.read(2))[0]
 
         # Payload
@@ -494,6 +561,21 @@ class TableMapEvent(BinLogEvent):
         self.packet.advance(1)
         self.table_length = byte2int(self.packet.read(1))
         self.table = self.packet.read(self.table_length).decode()
+
+        if self.__only_tables is not None and self.table not in self.__only_tables:
+            self._processed = False
+            return
+        elif self.__ignored_tables is not None and self.table in self.__ignored_tables:
+            self._processed = False
+            return
+
+        if self.__only_schemas is not None and self.schema not in self.__only_schemas:
+            self._processed = False
+            return
+        elif self.__ignored_schemas is not None and self.schema in self.__ignored_schemas:
+            self._processed = False
+            return
+
         self.packet.advance(1)
         self.column_count = self.packet.read_length_coded_binary()
 
@@ -504,14 +586,39 @@ class TableMapEvent(BinLogEvent):
         else:
             self.column_schemas = self._ctl_connection._get_table_information(self.schema, self.table)
 
-        # Read columns meta data
-        column_types = list(self.packet.read(self.column_count))
-        self.packet.read_length_coded_binary()
-        for i in range(0, len(column_types)):
-            column_type = column_types[i]
-            column_schema = self.column_schemas[i]
-            col = Column(byte2int(column_type), column_schema, from_packet)
-            self.columns.append(col)
+        ordinal_pos_loc = 0
+
+        if len(self.column_schemas) != 0:
+            # Read columns meta data
+            column_types = list(self.packet.read(self.column_count))
+            self.packet.read_length_coded_binary()
+            for i in range(0, len(column_types)):
+                column_type = column_types[i]
+                try:
+                    column_schema = self.column_schemas[ordinal_pos_loc]
+
+                    # only acknowledge the column definition if the iteration matches with ordinal position of
+                    # the column. this helps in maintaining support for restricted columnar access
+                    if i != (column_schema['ORDINAL_POSITION'] - 1):
+                        # raise IndexError to follow the workflow of dropping columns which are not matching the
+                        # underlying table schema
+                        raise IndexError
+
+                    ordinal_pos_loc += 1
+                except IndexError:
+                    # this a dirty hack to prevent row events containing columns which have been dropped prior
+                    # to pymysqlreplication start, but replayed from binlog from blowing up the service.
+                    # TODO: this does not address the issue if the column other than the last one is dropped
+                    column_schema = {
+                        'COLUMN_NAME': '__dropped_col_{i}__'.format(i=i),
+                        'COLLATION_NAME': None,
+                        'CHARACTER_SET_NAME': None,
+                        'COLUMN_COMMENT': None,
+                        'COLUMN_TYPE': 'BLOB',  # we don't know what it is, so let's not do anything with it.
+                        'COLUMN_KEY': '',
+                    }
+                col = Column(byte2int(column_type), column_schema, from_packet)
+                self.columns.append(col)
 
         self.table_obj = Table(self.column_schemas, self.table_id, self.schema,
                                self.table, self.columns)
